
(* Expressions de type. On les paramtre par le type ['ident] des
   identificateurs car ceux-ci seront localiss  l'issue du parsing
   (type [ident] ci-dessous) puis prendront le type [string]  l'issue
   du typage, pour permettre des comparaisons avec [=]. *)

type 'ident type_expr =
  | Tvoid
  | Tboolean
  | Tint
  | Tdouble
  | Tclass of 'ident
  | Tnull

(* Constantes *)

type constant = 
  | Cint of int
  | Cdouble of float
  | Cstring of (Location.t * string)
  | Cboolean of bool

(* Oprations unaires *)

type unop = 
  Upre_inc | Upost_inc | Upre_dec | Upost_dec | Unot | Uminus | Uplus |
  (* celles-ci apparaissent aprs le typage *)
  Udouble_of_int | Uint_of_double | 
  Uminus_f | Ustring_of_int | Ustring_of_float

(* Oprations binaires *)

type binop = 
  Beq | Bneq | Blt | Ble | Bgt | Bge | Badd | Bsub | Bmul | Bdiv | 
  Bmod | Band | Bor |
  (* celles-ci apparaissent aprs le typage *)
  Blt_f | Ble_f | Bgt_f | Bge_f | Badd_f | Bsub_f | Bmul_f | Bdiv_f | Badd_s

(* Information obtenue plus tard au typage, ncessaire pour produire le code *)

type class_info =
  | Cunknown
  | Cstatic of string (* champ ou mthode statique de C *)
  | Cdynamic of string * bool (* champ ou mthode dynamique de C / priv *)
  | Clocal (* variable locale *)

(* Expressions. Elles sont paramtres par le type des identificateurs, 
   de la mme manire que les types. *)

type 'ident expr = {
  expr_node : 'ident expr_node;
  expr_loc : Location.t
}

and 'ident expr_node =
  | Constant of constant
  | Null
  | This
  | Field of 'ident expr option * class_info * 'ident
  | Method_call of 'ident expr option * class_info * 'ident * 'ident expr list
  | New of 'ident * string option * 'ident expr list
  | Assign of 'ident expr * 'ident expr
  | Unop of unop * 'ident expr
  | Binop of binop * 'ident expr * 'ident expr
  | Cast of 'ident type_expr * 'ident expr
  | Instanceof of 'ident expr * 'ident type_expr
  (* apparait seulement aprs le typage *)
  | Println of 'ident expr

(* Modificateurs *)

type modifiers = { m_static : bool; m_private : bool }

(* Dclaration de variables (autant pour les variables d'instances que pour
   les variables locales) *)

type 'ident variable = { 
  variable_id : 'ident;
  variable_initializer : 'ident expr option;
}

type 'ident variables_decl = { 
  variables_modifiers : modifiers;
  variables_type : 'ident type_expr;
  variables_decls :  'ident variable list;
}

(* Instructions *)

type 'ident statement = {
  st_node : 'ident statement_node;
  st_loc : Location.t
}

and 'ident statement_node =
  | Skip
  | Expr of 'ident expr
  | Var of 'ident variables_decl
  | If of 'ident expr * 'ident statement * 'ident statement
  | For of 'ident statement list * 'ident expr * 
           'ident statement list * 'ident statement
  | While of 'ident expr * 'ident statement
  | Block of 'ident statement list
  | Return of 'ident expr option

(* Paramtres (arguments formels des constructeurs et mthodes) *)

type 'ident parameter = 'ident type_expr * 'ident

(* Dclaration de constructeur *)

type 'ident constr_decl = {
  constr_modifiers : modifiers;
  constr_name : 'ident;
  constr_parameters : 'ident parameter list;
  constr_body : 'ident statement list;
}

(* Dclaration de mthode *)

type 'ident method_decl = {
  method_modifiers : modifiers;
  method_type : 'ident type_expr;
  method_name : 'ident;
  method_parameters : 'ident parameter list;
  method_body : 'ident statement list;
}

(* Dclaration de classe *)

type 'ident field = 
  | Variable of 'ident variables_decl
  | Method of 'ident method_decl
  | Constructor of 'ident constr_decl

type 'ident class_decl = { 
  class_id : 'ident;
  class_extends : 'ident option;
  class_fields : 'ident field list;
}

(* Le rsultat du parsing est une liste de dclarations de classes,
   dans lesquelles les identificateurs sont localiss. *)

type ident = { id_name : string; id_loc : Location.t }

type compilation_unit = ident class_decl list

