
(* Production de code *)

open Format
open Ast

(* Allocation sur la pile (globaux) *)

let gp = ref 1 (* null prend la place 0 *)

let alloc f = let g = !gp in incr gp; f (); g
let alloc_gint fmt n = alloc (fun () -> fprintf fmt "\tpushi %d\n" n)
let alloc_gfloat fmt f = alloc (fun () -> fprintf fmt "\tpushf %f\n" f)
let alloc_ga fmt l = alloc (fun () -> fprintf fmt "\tpusha %s\n" l)
let alloc_g fmt ofs = alloc (fun () -> fprintf fmt "\tpushg %d\n" ofs)

let alloc_gtype fmt = function
  | Tvoid | Tint | Tboolean -> alloc_gint fmt 0
  | Tdouble -> alloc_gfloat fmt 0.0
  | Tclass _ | Tnull -> alloc_g fmt 0

(*s Tables *)

(* Table des mthodes dynamiques : donne leur offset par rapport au desc. *)

let (dynamic_methods : (string * string, int) Hashtbl.t) = Hashtbl.create 17

(* Table des variables dynamiques : donne leur offset par rapport  l'objet *)

let (dynamic_variables : (string * string, int) Hashtbl.t) = Hashtbl.create 17

(* Table des variables statiques : donne directement [n] pour [GP[n]] *)
let static_variables = Hashtbl.create 17

(* Table des descripteurs de classe *)

type desc = { 
  id : string; 
  where : int; (*r pos. par rapport  [GP] *)
  mutable dynamic_m : string list; (*r mthodes dynamiques *)
  mutable dynamic_v : string variable list; (*r variables dynamiques *)
  mutable size : int; (*r taille du bloc objet *)
}
let classes_desc = Hashtbl.create 17

let already_compiled = Hashtbl.mem classes_desc

let find_desc = Hashtbl.find classes_desc

(* La classe [Object] correspond au descripteur [0], qui n'a pas besoin d'tre
   allou car jamais consult *)

let object_desc = 
  { id = "Object"; where = 0; dynamic_m = []; dynamic_v = []; size = 1 }
let () = Hashtbl.add classes_desc "Object" object_desc

(* Ajout d'un descripteur de classe. *)

let add_desc fmt c super = 
  let n = alloc_gint fmt super in
  let d = { id = c; where = n; dynamic_m = []; dynamic_v = []; size = 1 } in
  Hashtbl.add classes_desc c d;
  d

(* Dclaration des mthodes dynamiques *)

let declare_dynamic_methods fmt desc super f =
  (* on collecte toutes les mthodes dynamiques de [C = desc.id] *)
  let all_dm = Hashtbl.create 17 in
  let add = function	    
    | Method md when not md.method_modifiers.m_static -> 
	Hashtbl.add all_dm md.method_name ()
    | _ -> ()
  in
  let () = List.iter add f in
  (* d'abord les mthodes commune avec la super-classe *)
  let declare m =
    let n = 
      (* [m] est-elle surcharge ? *)
      if Hashtbl.mem all_dm m then 
	alloc_ga fmt (desc.id ^ "_" ^ m)
      else
	alloc_g fmt (super.where + Hashtbl.find dynamic_methods (super.id, m))
    in
    Hashtbl.add dynamic_methods (desc.id, m) (n - desc.where);
    Hashtbl.remove all_dm m
  in    
  List.iter declare super.dynamic_m;
  desc.dynamic_m <- super.dynamic_m;
  (* ensuite les nouvelles mthodes *)
  let declare_new m _ =
    desc.dynamic_m <- desc.dynamic_m @ [m];
    let n = alloc_ga fmt (desc.id ^ "_" ^ m) in
    Hashtbl.add dynamic_methods (desc.id, m) (n - desc.where)
  in
  Hashtbl.iter declare_new all_dm

(* Dclaration des attributs dynamiques *)

let declare_dynamic_variables fmt desc super f =
  (* on collecte toutes les variables dynamiques de [C = desc.id] *)
  let all_dv = Hashtbl.create 17 in
  let add = function	    
    | Variable vd when not vd.variables_modifiers.m_static -> 
	List.iter (fun v -> Hashtbl.add all_dv v ()) vd.variables_decls
    | _ -> ()
  in
  let () = List.iter add f in
  (* d'abord les variables commune avec la super-classe *)
  let declare v =
    let ofs = Hashtbl.find dynamic_variables (super.id, v.variable_id) in
    Hashtbl.add dynamic_variables (desc.id, v.variable_id) ofs
  in    
  List.iter declare super.dynamic_v;
  desc.dynamic_v <- super.dynamic_v;
  (* ensuite les nouvelles variables *)
  let free = ref (1 + List.length super.dynamic_v) in
  let declare_new v _ =
    desc.dynamic_v <- desc.dynamic_v @ [v];
    let ofs = !free in
    incr free;
    Hashtbl.add dynamic_variables (desc.id, v.variable_id) ofs
  in
  Hashtbl.iter declare_new all_dv;
  desc.size <- 1 + List.length desc.dynamic_v

(* Dclaration des attributs statiques *)

let declare_static_variables fmt desc super f =
  (* pour commencer, on hrite de toutes les variables statiques de super *)
  Hashtbl.iter 
    (fun (c,x) n -> 
       if c = super.id then Hashtbl.add static_variables (desc.id, x) n) 
    static_variables;
  (* ensuite on alloue les variables statiques propres *)
  let declare_variables = function
    | Variable vd when vd.variables_modifiers.m_static ->
	let declare_variable v = 
	  let n = alloc_gtype fmt vd.variables_type in
	  Hashtbl.add static_variables (desc.id, v.variable_id) n
	in
	List.iter declare_variable vd.variables_decls
    | Constructor _ | Variable _ | Method _ ->
	()
  in
  List.iter declare_variables f

(* On remplit les tables en suivant un tri topologique des classes *)

let to_compile = Hashtbl.create 17

let rec compile_desc fmt c =
  if already_compiled c then 
    find_desc c
  else begin
    let cd = Hashtbl.find to_compile c in
    let super = match cd.class_extends with 
      | Some c' -> compile_desc fmt c'
      | None -> object_desc
    in
    fprintf fmt "// descripteur classe %s\n" c;
    let desc = add_desc fmt c super.where in
    declare_dynamic_methods fmt desc super cd.class_fields;
    declare_static_variables fmt desc super cd.class_fields;
    declare_dynamic_variables fmt desc super cd.class_fields;
    desc
  end

(*s Compilation du code proprement dit *)

(* Un contexte donne la position par rapport  [FP] des variables locales,
   y compris [this] et [%return%], qui est la variable stockant la valeur
   de retour d'une fonction. *)

module Context = struct

  module Smap = Map.Make(String)

  (* couple [(i,m)] o [i] indique le premier emplacement libre au dessus de
    [FP] (pour la prochaine variable locale), et o la table [m] donne l'offset
    par rapport  [FP] de chaque variable locale *)

  type t = int * int Smap.t

  let empty = 0, Smap.empty

  let find x (_,m) = Smap.find x m

  (* Cre un contexte d'appel, contenant "%return%", "this" (si [static] 
     vaut vrai) et les paramtres [params] *)

  let call params static = 
    let n = List.length params in
    let ctx,i = 
      List.fold_right (fun (_,x) (ctx,i) -> Smap.add x i ctx, i-1) 
	params (Smap.empty, -1)
    in
    let ctx,i = if static then ctx,i else Smap.add "this" i ctx, i-1 in
    0, Smap.add "%return%" i ctx

  (* Ajout de variables locales (au dessus de [FP]) *)

  let add_locals { variables_decls = vl } ctx =
    List.fold_right 
      (fun {variable_id=x} (i,m) -> i + 1, Smap.add x i m)
      vl ctx

end

(* Compte le nombre maximum de variables locales le long de tous les
   flots d'excution possibles d'un bloc. *)

let count_locals bl =
  let rec count_block = function
    | [] -> 0
    | i :: bl -> count i + count_block bl
  and count st = match st.st_node with
    | Skip | Expr _ | Return _ -> 0
    | Var { variables_decls = vl } -> List.length vl
    | If (_, i1, i2) -> max (count i1) (count i2)
    | For (bl1, _, bl2, i3) -> count_block bl1 + count_block bl2 + count i3
    | While (_, i) -> count i
    | Block bl -> count_block bl
  in 
  count_block bl

(* Le type [result] suivant indique que faire du rsultat de l'valuation
   d'une expression : l'ignorer ([Discard]), le laisser en sommet de pile 
   ([Stack]) ou le stocker quelque part ([Store]). *)

type where =
  | Fp of int (* position relative par rapport  [FP] *)
  | Gp of int (* position relative par rapport  [GP] *)

type result = 
  | Discard
  | Stack (* au sommet de la pile *)
  | Store of where

let move fmt = function
  | Discard -> fprintf fmt "\tpop 1\n"
  | Stack -> ()
  | Store (Fp n) -> fprintf fmt "\tstorel %d\n" n
  | Store (Gp n) -> fprintf fmt "\tstoreg %d\n" n

(* Oprations binaires *)

let compile_binop = function
  | Beq -> "equal"
  | Blt -> "inf" 
  | Ble -> "infeq"
  | Bgt -> "sup"
  | Bge -> "supeq"
  | Badd -> "add"
  | Bsub -> "sub"
  | Bmul -> "mul"
  | Bdiv -> "div"
  | Bmod -> "mod"
  | Blt_f -> "finf" 
  | Ble_f -> "finfeq"
  | Bgt_f -> "fsup"
  | Bge_f -> "fsupeq"
  | Badd_f -> "fadd"
  | Bsub_f -> "fsub"
  | Bmul_f -> "fmul"
  | Bdiv_f -> "fdiv"
  (* celles-ci seront traites spcifiquement *)
  | Badd_s | Bneq | Band | Bor -> assert false

(* Gnration d'tiquettes fraches *)

let new_label = let r = ref 0 in fun () -> incr r; "_label_" ^ string_of_int !r

(* Pour la construction d'expressions pendant la compilation *)

let dummy e = { expr_node = e ; expr_loc = Location.dummy }

(* Valeur par dfaut en Java *)

let compile_default_value fmt = function
  | Tvoid | Tint | Tboolean -> fprintf fmt "\tpushi 0\n"
  | Tdouble -> fprintf fmt "\tpushf 0.0\n"
  | Tclass _ | Tnull -> fprintf fmt "\tpushg 0\n"

(* Le descripteur d'une classe *)

let desc_of_type ty =
  let c = match ty with Tclass c -> c | _ -> assert false in 
  (find_desc c).where

(* Compilation des expressions *)

let rec compile_expr fmt ctx res e = 
  compile_expr_node fmt ctx res e.expr_node

and compile_expr_node fmt ctx res = function
  | Null | This | Constant _ when res = Discard -> (* OPTIM *)
      ()
  | Constant (Cint n) ->
      fprintf fmt "\tpushi %d\n" n; move fmt res
  | Constant (Cdouble f) ->
      fprintf fmt "\tpushf %f\n" f; move fmt res
  (* convention de la MV : 0 reprsente faux et 1 vrai *)
  | Constant (Cboolean false) ->
      fprintf fmt "\tpushi 0\n"; move fmt res
  | Constant (Cboolean true) ->
      fprintf fmt "\tpushi 1\n"; move fmt res
  | Constant (Cstring (_, s)) ->
      let d = find_desc "String" in
      (* on alloue un bloc (this) *)
      fprintf fmt "\talloc 2 // \"%s\"\n" s;
      (* on met le descripteur de String dans le champ 0 *)
      fprintf fmt "\tdup 1\n\tpushi %d\n\tstore 0\n" d.where;
      (* on met l'adresse de la chane dans le champ 1 *)
      fprintf fmt "\tdup 1\n\tpushs \"%s\"\n\tstore 1\n" s;
      move fmt res
  | Null ->
      fprintf fmt "\tpushg 0  // null \n"; move fmt res
  | This ->
      let n = Context.find "this" ctx in
      fprintf fmt "\tpushl %d  // this\n" n
  | Field (None, Clocal, x) ->
      let ofs = Context.find x ctx in
      fprintf fmt "\tpushl %d  // %s\n" ofs x; 
      move fmt res (* OPTIM selon res ? *)
  | Field (_, Clocal, _) -> 
      assert false
  | Field (e, Cstatic c, x) ->
      (match e with None -> () | Some e -> compile_expr fmt ctx Discard e);
      let n = Hashtbl.find static_variables (c, x) in
      fprintf fmt "\tpushg %d  // %s.%s\n" n c x;
      move fmt res
  | Field (o, Cdynamic (c,_), x) ->
      (match o with
	 | None -> compile_expr_node fmt ctx Stack This
	 | Some e -> compile_expr fmt ctx Stack e);
      let ofs = Hashtbl.find dynamic_variables (c, x) in
      fprintf fmt "\tload %d  // %s\n" ofs x;
      move fmt res
  | Field _ ->
      assert false
  | Method_call (_, (Cunknown | Clocal), _, _) ->
      assert false
  | Method_call (Some o, Cdynamic ("String",_), "equals_String", [e]) ->
      compile_expr fmt ctx Stack o;
      fprintf fmt "\tload 1\n";
      compile_expr fmt ctx Stack e;
      fprintf fmt "\tload 1\n\tequal\n";
      move fmt res
  | Method_call (o, Cstatic c, m, el) ->
      (match o with None -> () | Some e -> compile_expr fmt ctx Discard e);
      (* place pour le rsultat *)
      fprintf fmt "\tpushn 1 // pour le rsultat de l'appel\n";
      (* les arguments *)
      compile_expr_l fmt ctx el;
      (* l'appel *)
      fprintf fmt "\tpusha %s_%s\n" c m;
      fprintf fmt "\tcall\n";
      (* nettoyage des arguments *)
      let n = List.length el in if n > 0 then fprintf fmt "\tpop %d\n" n;
      move fmt res
  | Method_call (o, Cdynamic (c,pr), m, el) ->
      (* place pour le rsultat *)
      fprintf fmt "\tpushn 1 // pour le rsultat de l'appel\n";
      (* l'objet this *)
      (match o with
	 | None -> compile_expr_node fmt ctx Stack This
	 | Some e -> compile_expr fmt ctx Stack e);
      (* les arguments *)
      compile_expr_l fmt ctx el;
      let n = List.length el in
      (* l'adresse  appeler *)
      begin 
	if pr then
	  (* appel statique car mthode prive *)
	  fprintf fmt "\tpusha %s_%s // private\n" c m
	else begin
	  (* appel dynamique *)
	  fprintf fmt "\tpushgp\n";
	  (* on accde au descripteur de classe (champ 0 de l'objet) *)
	  fprintf fmt "\tpushsp\n\tload %d\n\tload 0\n" (-n-2); 
	  (* on lui ajoute l'offset pour la mthode m *)
	  let ofs = Hashtbl.find dynamic_methods (c, m) in
	  fprintf fmt "\tpushi %d\n\tadd\n\tloadn\n" ofs
	end
      end;
      fprintf fmt "\tcall // %s.%s dynamic\n" c m;
      (* nettoyage des arguments et de this *)
      if n >= 0 then fprintf fmt "\tpop %d\n" (n + 1);
      move fmt res
  | New (c, Some m, el) ->
      let d = find_desc c in
      (* on alloue un bloc (this) *)
      fprintf fmt "\talloc %d // new %s\n" d.size c;
      (* on met le descripteur de C dans le champ 0 *)
      fprintf fmt "\tdup 1\n\tpushi %d\n\tstore 0\n" d.where;
      (* on initialise les variables dynamiques *)
      fprintf fmt "\tpusha _init_%s\n\tcall\n" c;
      (* on empile les arguments *)
      compile_expr_l fmt ctx el;
      let n = List.length el in
      (* on appelle le constructeur *)
      fprintf fmt "\tpusha %s_%s\n" c m;
      fprintf fmt "\tcall\n";
      (* on nettoie les arguments, this reste et c'est justement le rsultat *)
      if n > 0 then fprintf fmt "\tpop %d\n" n;
      move fmt res
  | New (_, None, _) ->
      assert false
  (* OPTIM: affectation d'une variable locale *)
  | Assign ({ expr_node = Field (None, Clocal, x) }, e) ->
      let n = Context.find x ctx in
      compile_expr fmt ctx Stack e;
      (match res with
	 | Discard -> fprintf fmt "\tstorel %d  // %s\n" n x
	 | _ -> fprintf fmt "\tdup 1\n\tstorel %d  // %s\n" n x; move fmt res)
  (* cas gnral d'affectation *)
  | Assign (e1, e2) ->
      (* ordre valuation : e1 avant e2 *)
      (* 1. on rserve de la place pour v2 *)
      fprintf fmt "\tpushn 1\n";
      (* 2. on compile la valeur gauche e1 *)
      compile_left_value fmt ctx e1;
      (* 3. on compile e2  *)
      compile_expr fmt ctx Stack e2;
      (match res with
	 | Discard -> 
	     fprintf fmt "\tstoren\n\tpop 1\n"
	 | _ -> 
	     (* on copie v2 en SP[-5] et on affecte *)
	     fprintf fmt "\tdup 1\n\tpushsp\n\tswap\n\tstore -5\n\tstoren\n"; 
	     move fmt res)
  | Unop (Uplus, _) -> assert false
  | Unop (Uminus, e) ->
      (* 0 - e *)
      compile_expr_node fmt ctx Stack (Constant (Cint 0));
      compile_expr fmt ctx Stack e;
      fprintf fmt "\tsub\n"; move fmt res
  | Unop (Uminus_f, e) ->
      (* 0.0 - e *)
      compile_expr_node fmt ctx Stack (Constant (Cdouble 0.0));
      compile_expr fmt ctx Stack e;
      fprintf fmt "\tfsub\n"; move fmt res
  | Unop (Unot, e) ->
      compile_expr fmt ctx Stack e; fprintf fmt "\tnot\n"; move fmt res
  | Unop (Udouble_of_int, e) ->
      compile_expr fmt ctx Stack e; fprintf fmt "\titof\n"; move fmt res
  | Unop (Uint_of_double, e) ->
      compile_expr fmt ctx Stack e; fprintf fmt "\tftoi\n"; move fmt res
  | Unop (Ustring_of_int | Ustring_of_float as op, e) ->
      let d = find_desc "String" in
      fprintf fmt "\talloc 2\n";
      fprintf fmt "\tdup 1\n\tpushi %d\n\tstore 0\n\tdup 1\n" d.where;
      compile_expr fmt ctx Stack e;
      let op = match op with Ustring_of_int -> "stri" | _ -> "strf" in
      fprintf fmt "\t%s\n\tstore 1\n" op;
      move fmt res
  | Unop ((Upost_dec | Upre_dec | Upost_inc | Upre_inc as op), e) ->
      (* TODO: OPTIM possible quand res = Discard, quand variable locale *)
      fprintf fmt "\tpushn 1\n";
      compile_left_value fmt ctx e;
      fprintf fmt "\tdup 2\n\tloadn\n";
      (match op with 
	 | Upost_dec | Upost_inc -> 
	     fprintf fmt "\tdup 1\n\tpushsp\n\tswap\n\tstore -5\n"
	 | _ -> ());
      (match op with 
	 | Upost_inc | Upre_inc -> fprintf fmt "\tpushi 1\n\tadd\n"
	 | _ -> fprintf fmt "\tpushi 1\n\tsub\n");
      (match op with 
	 | Upre_dec | Upre_inc -> 
	     fprintf fmt "\tdup 1\n\tpushsp\n\tswap\n\tstore -5\n"
	 | _ -> ());
      fprintf fmt "\tstoren\n";
      move fmt res
  | Binop (Bneq, e1, e2) ->
      compile_expr fmt ctx Stack e1;
      compile_expr fmt ctx Stack e2;
      fprintf fmt "\tequal\n\tnot\n"; move fmt res
  | Binop (Band, e1, e2) ->
      compile_expr fmt ctx Stack e1;
      let l = new_label () in
      fprintf fmt "\tdup 1\n\tjz %s\n\tpop 1\n" l;
      compile_expr fmt ctx Stack e2;
      fprintf fmt "%s:\n" l; move fmt res
  | Binop (Bor, e1, e2) ->
      compile_expr fmt ctx Stack e1;
      let l = new_label () in
      fprintf fmt "\tdup 1\n\tnot\n\tjz %s\n\tpop 1\n" l;
      compile_expr fmt ctx Stack e2;
      fprintf fmt "%s:\n" l; move fmt res
  | Binop (Badd_s, e1, e2) ->
      let d = find_desc "String" in
      fprintf fmt "\talloc 2\n";
      fprintf fmt "\tdup 1\n\tpushi %d\n\tstore 0\n\tdup 1\n" d.where;
      compile_expr fmt ctx Stack e1;
      fprintf fmt "\tload 1\n";
      compile_expr fmt ctx Stack e2;
      fprintf fmt "\tload 1\n";
      fprintf fmt "\tswap\nconcat\n\tstore 1\n"; move fmt res
  | Binop (op, e1, e2) ->
      compile_expr fmt ctx Stack e1;
      compile_expr fmt ctx Stack e2;
      fprintf fmt "\t%s\n" (compile_binop op); move fmt res
  | Cast (ty, e) ->
      compile_expr fmt ctx Stack e;
      let l = new_label () in
      fprintf fmt "\tdup 1\n\tpushg 0\n\tequal\n\tjz %s\n" l;
      fprintf fmt "\terr \"cast failure (null)\"\n";
      fprintf fmt "%s:\n\tdup 1\n\tload 0\n" l;
      fprintf fmt "\tpushi %d\n" (desc_of_type ty);
      fprintf fmt "\tpusha _cast_\n\tcall\n\tpop 2\n";
      move fmt res
  | Instanceof(e, ty) ->
      fprintf fmt "\tpushn 1\n";
      compile_expr fmt ctx Stack e;
      let l = new_label () in
      let done_ = new_label () in
      fprintf fmt "\tdup 1\n\tpushg 0\n\tequal\n\tjz %s\n" l;
      fprintf fmt "\tpop 1\n\tjump %s\n" done_;
      fprintf fmt "%s:\n\tload 0\n" l;
      fprintf fmt "\tpushi %d\n" (desc_of_type ty);
      fprintf fmt "\tpusha _instanceof_\n\tcall\n\tpop 2\n";
      fprintf fmt "%s:\n" done_;
      move fmt res
  | Println e ->
      compile_expr fmt ctx Stack e;
      fprintf fmt "\tload 1\n\twrites\n"

and compile_expr_l fmt ctx = function
  | [] -> 
      ()
  | e :: el ->
      compile_expr fmt ctx Stack e;
      compile_expr_l fmt ctx el

(* Compilation des valeurs gauches : empile une adresse et un offset *)

and compile_left_value fmt ctx e = 
  compile_left_value_node fmt ctx e.expr_node

and compile_left_value_node fmt ctx = function 
  | Field (None, Clocal, x) ->
      let n = Context.find x ctx in
      fprintf fmt "\tpushfp\n\tpushi %d  // %s\n" n x
  | Field (_, Cstatic c, x) ->
      let n = Hashtbl.find static_variables (c, x) in
      fprintf fmt "\tpushgp\n\tpushi %d  // %s.%s\n" n c x
  | Field (o, Cdynamic (c,_), x) ->
      (match o with
	 | None -> compile_expr_node fmt ctx Stack This
	 | Some e -> compile_expr fmt ctx Stack e);
      let ofs = Hashtbl.find dynamic_variables (c, x) in
      fprintf fmt "\tpushi %d  // %s\n" ofs x
  | _ ->
      assert false

(* Compilation des instructions *)

let rec compile_statement fmt ctx st = 
  compile_statement_node fmt ctx st.st_node

and compile_statement_node fmt ctx = function
  | Skip -> 
      ctx
  | Expr e ->
      compile_expr fmt ctx Discard e;
      ctx
  | Return None ->
      fprintf fmt "\treturn\n";
      ctx
  | Return (Some e) ->
      compile_expr fmt ctx (Store (Fp (Context.find "%return%" ctx))) e;
      fprintf fmt "\treturn\n";
      ctx
  | Var vd ->
      let ctx' = Context.add_locals vd ctx in
      let init_var { variable_id = x; variable_initializer = e } = match e with
	| None -> 
	    ()
	| Some e -> 
	    let vx = dummy (Field (None, Clocal, x)) in
	    compile_expr fmt ctx' Discard (dummy (Assign (vx, e)))
      in
      List.iter init_var vd.variables_decls;
      ctx'
  | If (e, i1, i2) ->
      compile_expr fmt ctx Stack e;
      let l1 = new_label () in
      let l2 = new_label () in
      fprintf fmt "\tjz %s\n" l1;
      let _ = compile_statement fmt ctx i1 in
      fprintf fmt "\tjump %s\n%s:\n" l2 l1;
      let _ = compile_statement fmt ctx i2 in
      fprintf fmt "%s:\n" l2;
      ctx
  | For (bl1, e, bl2, i3) ->
      let ctx' = compile_block fmt ctx bl1 in
      let l1 = new_label () in
      let l2 = new_label () in
      fprintf fmt "%s:\n" l1;
      compile_expr fmt ctx' Stack e;
      fprintf fmt "\tjz %s\n" l2;
      let _ = compile_statement fmt ctx' i3 in
      let _ = compile_block fmt ctx' bl2 in
      fprintf fmt "\tjump %s\n%s:\n" l1 l2;
      ctx
  | While (e, i) ->
      let l1 = new_label () in
      let l2 = new_label () in
      fprintf fmt "%s:\n" l1;
      compile_expr fmt ctx Stack e;
      fprintf fmt "\tjz %s\n" l2;
      let _ = compile_statement fmt ctx i in
      fprintf fmt "\tjump %s\n%s:\n" l1 l2;
      ctx
  | Block bl ->
      compile_block fmt ctx bl

and compile_block fmt ctx = function
  | [] -> 
      ctx
  | i :: bl ->
      let ctx' = compile_statement fmt ctx i in
      compile_block fmt ctx' bl

(* Compilation des fonctions d'une classe 
   ( la diffrence du remplissage des tables
   fait en premier lieu, on peut compiler les classes dans un ordre
   arbitraire, ici celui du fichier) *)

let compile_class fmt cd =
  let c = cd.class_id in
  fprintf fmt "// code classe %s\n" c;
  (* fonction d'initialisation des variables dynamiques *)
  fprintf fmt "_init_%s:\n" c;
  (match cd.class_extends with
     | None -> ()
     | Some s -> fprintf fmt "\tpusha _init_%s\n\tcall\n" s);
  let ctx = Context.call [] false in
  let compile_init_var ty v = 
    let ev = Field (None, Cdynamic (c,false), v.variable_id) in
    compile_left_value_node fmt ctx ev;
    (match v.variable_initializer with
       | None -> compile_default_value fmt ty
       | Some e -> compile_expr fmt ctx Stack e);
    fprintf fmt "\tstoren\n"
  in
  let compile_init = function
    | Variable vd when not vd.variables_modifiers.m_static -> 
	List.iter (compile_init_var vd.variables_type) vd.variables_decls 
    | _ ->
	()
  in
  List.iter compile_init cd.class_fields;
  fprintf fmt "\treturn\n";
  (* code des constructeurs et mthodes *)
  let alloc_locals b = 
    let n = count_locals b in
    if n > 0 then fprintf fmt "\tpushn %d\n" n
  in
  let compile_method = function
    | Method md ->
	fprintf fmt "%s_%s:\n" c md.method_name;
	let ctx = 
	  Context.call md.method_parameters md.method_modifiers.m_static
	in
	alloc_locals md.method_body;
	let _ = compile_block fmt ctx md.method_body in
	(* si le type de retour n'est pas [void], il y a forcment un [return]
	   dans le flot de contrle *)
	(match md.method_type with
	   | Tvoid -> 
	       fprintf fmt "\treturn\n" 
	       (* en fait, inutile s'il y en a dj un dans le flot de contrle
		  (vrifi au typage) *)
	   | _ -> ())
    | Constructor cd ->
	fprintf fmt "%s:\n" cd.constr_name;
	let ctx = 
	  Context.call cd.constr_parameters cd.constr_modifiers.m_static
	in
	alloc_locals cd.constr_body;
	let _ = compile_block fmt ctx cd.constr_body in
	fprintf fmt "\treturn\n"
    | Variable _ -> 
	()
  in
  List.iter compile_method cd.class_fields

(* Initialisation des champs statiques: on les fait suivant l'ordre 
   d'apparition dans le fichier *)

let init_static_fields fmt p =
  let init_var c v = match v.variable_initializer with
    | Some e ->
	let x = v.variable_id in
	fprintf fmt "// initialisation de %s.%s\n" c x;
	compile_expr fmt Context.empty Stack e;
	let n = Hashtbl.find static_variables (c, x) in
	fprintf fmt "\tstoreg %d\n" n
    | None ->
	()
  in
  let init_class c = function
    | Variable vd when vd.variables_modifiers.m_static ->
	List.iter (init_var c) vd.variables_decls
    | _ -> 
	()
  in
  List.iter (fun cd -> List.iter (init_class cd.class_id) cd.class_fields) p

(* Fonctions utilitaires ralisant respectivement le transtypage
   et l'opration [instanceof]. *)

let cast_function fmt = 
  fprintf fmt "_cast_:\n";
  let start = new_label () in
  let next = new_label () in
  let done_ = new_label () in
  fprintf fmt "%s:\n" start;
  fprintf fmt "\tdup 2\n\tequal\n\tnot\n\tjz %s\n" done_;
  fprintf fmt "\tpushl -2\n\tpushi 0\n\tequal\n\tjz %s\n" next;
  fprintf fmt "\terr \"cast failure\"\n";
  fprintf fmt "%s:\n" next;
  fprintf fmt "\tpushgp\n\tpushl -2\n\tloadn\n\tstorel -2\n\tjump %s\n" start;
  fprintf fmt "%s:\n\treturn\n" done_

let instanceof fmt = 
  fprintf fmt "_instanceof_:\n";
  let start = new_label () in
  let next = new_label () in
  let done_ = new_label () in
  fprintf fmt "%s:\n" start;
  fprintf fmt "\tdup 2\n\tequal\n\tnot\n\tjz %s\n" done_;
  fprintf fmt "\tpushl -2\n\tpushi 0\n\tequal\n\tjz %s\n" next;
  fprintf fmt "\tpushi 0\n\tstorel -3\n\treturn\n";
  fprintf fmt "%s:\n" next;
  fprintf fmt "\tpushgp\n\tpushl -2\n\tloadn\n\tstorel -2\n\tjump %s\n" start;
  fprintf fmt "%s:\n\tpushi 1\n\tstorel -3\n\treturn\n" done_

let predefined_functions fmt = 
  cast_function fmt;
  instanceof fmt;
  fprintf fmt "_init_Object:\n\treturn\n"

(* Compilation d'un fichier *)

let call c m = dummy (Method_call (None, Cstatic c, m ^ "_", []))

let compile_file fmt p entry =
  fprintf fmt "// null stock en GP[0]\n\talloc 0\n";
  fprintf fmt "// descripteur classe String\n";
  let _ = add_desc fmt "String" 0 in
  List.iter (fun cd -> Hashtbl.add to_compile cd.class_id cd) p;
  Hashtbl.iter (fun c _ -> ignore (compile_desc fmt c)) to_compile;
  fprintf fmt "\tstart\n";
  init_static_fields fmt p;
  let (c,m) = entry in 
  fprintf fmt "// point d'entre = %s.%s\n" c m;
  compile_expr fmt Context.empty Discard (call c m);
  fprintf fmt "\tstop\n";
  List.iter (compile_class fmt) p;
  predefined_functions fmt

let compile file p entry =
  Typing.check_entry_point entry;
  let vm = (Filename.chop_suffix file ".java") ^ ".vm" in
  let c = open_out vm in
  let fmt = formatter_of_out_channel c in
  compile_file fmt p entry;
  fprintf fmt "@?";
  close_out c
