
(* Lexeur pour le compilateur Java *)

{ 

  open Lexing 
  open Ast
  open Parser

  (* Erreurs lexicales *)

  exception Lexical_error of string

  (* Tampon pour les chanes de caractres *)

  let string_buffer = Buffer.create 512

  (* Table des mots cls *)

  let kw_table = 
    let table = Hashtbl.create 97 in
    List.iter
      (fun (s,t) -> Hashtbl.add table s t)
      [ "boolean", BOOLEAN;
	"class", CLASS;
	"double", DOUBLE;
	"else", ELSE;
	"extends", EXTENDS;
	"false", FALSE;
	"for", FOR;
	"if", IF;
	"instanceof", INSTANCEOF;
	"int", INT;
	"new", NEW;
	"null", NULL;
	"private", PRIVATE;
	"return", RETURN;
	"static", STATIC;
	"this", THIS;
	"true", TRUE;
	"void", VOID;
	"while", WHILE;	
      ];
    table

  let ident_or_keyword s =
    try
      Hashtbl.find kw_table s
    with
      | Not_found -> IDENT s

}


(* Point d'entre du lexeur *)

rule token = parse
  | [' ' '\t' '\r' '\n']+ 
    { token lexbuf }
  | "/*" 
    { comment lexbuf; token lexbuf }
  | "//" [^ '\n']* '\n' 
    { token lexbuf }
  | ['A'-'Z''a'-'z''_'] ['A'-'Z''a'-'z''0'-'9''_']* 
      { ident_or_keyword (lexeme lexbuf) }
  | ';'
      { SEMICOLON }
  | ','
      { COMMA }
  | '.'
      { DOT }
  | '+'                 
      { PLUS }
  | '-'       
      { MINUS }
  | "++"                 
      { PLUSPLUS }
  | "--"       
      { MINUSMINUS }
  | '*'
      { STAR }
  | '/'
      { SLASH }
  | '%'
      { PERCENT }
  | "&&"
      { AMPERSANDAMPERSAND }
  | "||"
      { VERTICALBARVERTICALBAR }
  | "!"
      { BANG }
  | "?"
      { QUESTIONMARK }
  | ":"
      { COLON }
  | "=" 
      { EQ }
  | ">" 
      { COMP Bgt }
  | ">=" 
      { COMP Bge }
  | "<" 
      { COMP Blt }
  | "<=" 
      { COMP Ble }
  | "==" 
      { EQOP Beq }
  | "!="
      { EQOP Bneq }
  (* constantes entires *)
  | '0' | ['1'-'9']['0'-'9']*
      { INTEGER (int_of_string (lexeme lexbuf)) }
  (* constantes flotantes *)
  | ['0'-'9']+ '.' ['0'-'9']* (['e''E']['-''+']?['0'-'9']+)? 
  | '.' ['0'-'9']+ (['e''E']['-''+']?['0'-'9']+)? 
  | ['0'-'9']+ ['e''E']['-''+']?['0'-'9']+ 
      { REAL (float_of_string (lexeme lexbuf)) }
  | '('
      { LEFTPAR }
  | ')'
      { RIGHTPAR }
  | '{'
      { LEFTBRACE }
  | '}'
      { RIGHTBRACE }
  | '"'
      { let start = Lexing.lexeme_start lexbuf in
	Buffer.clear string_buffer;
	let eend = string lexbuf in
	STRING ((start,eend), Buffer.contents string_buffer) }
  | _ 
      { raise (Lexical_error (lexeme lexbuf)) }
  | eof
      { EOF }

(* Lecture d'une chane de caractres *)

and string = parse
  | '"'
      { lexeme_end lexbuf }
  | '\\' 'n' 
      { Buffer.add_string string_buffer "\\n";
	string lexbuf }
  | '\\' '\\' 
      { Buffer.add_string string_buffer "\\\\";
	string lexbuf }
  | '\\' '"' 
      { Buffer.add_string string_buffer "\\\"";
	string lexbuf }
  | [^ '\\' '"' '\n']+ 
      { Buffer.add_string string_buffer (lexeme lexbuf);
	string lexbuf }
  | '\\' 
      { raise (Lexical_error "illegal escape character") }
  | '\n' | eof
      { raise (Lexical_error "unterminated string") }

(* Lecture d'un commentaire *)

and comment = parse
  | "*/" { () }
  | _    { comment lexbuf }
  | eof  { raise (Lexical_error "unterminated comment") }


