
/* Parseur pour le compilateur Java */

%{

  open Ast

  let no_modifiers = { m_static = false; m_private = false }

  (* Construction d'expressions et d'instructions localises *)

  let loc () = symbol_start (), symbol_end ()

  let mk_expr e = { expr_node = e; expr_loc = loc () }

  let mk_expr_i i e = { expr_node = e; expr_loc = rhs_start i, rhs_end i }

  let mk_st s = { st_node = s; st_loc = loc () }

  let mk_st_i i s = { st_node = s; st_loc = rhs_start i, rhs_end i }

%}

%token <string> IDENT
%token <int> INTEGER
%token <float> REAL
%token <Location.t * string> STRING

/* Mots cls */

%token TRUE FALSE NULL THIS 
%token NEW 
%token BOOLEAN 
%token CLASS DOUBLE ELSE EXTENDS
%token FOR IF
%token INSTANCEOF INT 
%token PRIVATE
%token RETURN STATIC
%token VOID
%token WHILE

/* Symboles */

%token LEFTPAR 
%token RIGHTPAR LEFTBRACE RIGHTBRACE 
%token SEMICOLON COLON COMMA QUESTIONMARK DOT
%token EOF

/* Oprateurs */

%token EQ 
%token VERTICALBARVERTICALBAR
%token AMPERSANDAMPERSAND  
%token <Ast.binop> EQOP
%token <Ast.binop> COMP 
%token PLUS MINUS 
%token STAR SLASH PERCENT
%token PLUSPLUS MINUSMINUS BANG

/*s Prcdences */

%nonassoc THEN
%nonassoc ELSE

%right EQ 
%left VERTICALBARVERTICALBAR  /*r conditional OR ["||"] */
%left AMPERSANDAMPERSAND  /*r conditional AND ["&&"] */
%left EQOP                /*r ["=="] and ["!="] */
%left COMP INSTANCEOF  /*r ["<"], ["<="], [">"], [">="] and ["instanceof"] */
%left PLUS MINUS             /*r ["+"] and ["-"] */
%left STAR SLASH PERCENT     /*r ["*"], ["/"] and ["%"] */
%right UMINUS UPLUS PLUSPLUS MINUSMINUS BANG CAST
           /*r unary ["+"], ["-"], ["++"], ["--"], ["~"], ["!"] and cast */
%nonassoc par_expr
%left DOT

/*s Point d'entre */

%start compilation_unit
%type <Ast.compilation_unit> compilation_unit

%%

compilation_unit:
| class_declarations EOF { $1 }
;

/* dclaration de classe */

class_declarations:
| /* epsilon */ { [] }
| class_declaration class_declarations { $1 :: $2 }
;

class_declaration:
| CLASS ident extends_decl LEFTBRACE field_declarations RIGHTBRACE
  { { class_id = $2; class_extends = $3; class_fields = $5 } }
;

extends_decl:
| /* epsilon */
    { None }
| EXTENDS ident
    { Some $2 }
;

field_declarations:
| /* $\varepsilon$ */
    { [] }
| field_declaration field_declarations
    { $1::$2 }
;

field_declaration:
| method_declaration 
    { Method $1 }
| constructor_declaration 
    { Constructor $1 }
| variable_declaration 
    { Variable $1 }
;

/* dclaration de variables */

variable_declaration:
| modifiers type_expr variable_declarators SEMICOLON
    { { variables_modifiers = $1 ;
	variables_type = $2 ;
	variables_decls = $3 } } 
;

variable_declarators:
| variable_declarator
    { [$1] }
| variable_declarator COMMA variable_declarators
    { $1 :: $3 }
;

variable_declarator:
| ident
    { { variable_id = $1; variable_initializer = None } }
| ident EQ expr
    { { variable_id = $1; variable_initializer = Some $3 } }
;

/* dclaration de constructeur */

constructor_declaration:
| modifiers ident method_parameters LEFTBRACE statements RIGHTBRACE
    { { constr_modifiers = $1 ;
	constr_name = $2 ;
	constr_parameters = $3 ;
	constr_body = $5 } }
;

/* dclaration de mthode */

method_declaration:
| modifiers type_expr ident method_parameters LEFTBRACE statements RIGHTBRACE
  { { method_modifiers = $1;
      method_type = $2;
      method_name = $3;
      method_parameters = $4;
      method_body = $6 } }
| modifiers VOID ident method_parameters LEFTBRACE statements RIGHTBRACE
  { { method_modifiers = $1;
      method_type = Tvoid;
      method_name = $3;
      method_parameters = $4;
      method_body = $6 } }
;

/* arguments */

argument_list:
| /* epsilon */
    { [] }
| expr_comma_list
    { $1 }
;

expr:
| left_value EQ expr
    { mk_expr (Assign ($1, $3)) }
| MINUSMINUS expr 
    { mk_expr (Unop (Upre_dec, $2)) }
| PLUSPLUS expr 
    { mk_expr (Unop (Upre_inc, $2)) }
| expr PLUSPLUS 
    { mk_expr (Unop (Upost_inc, $1)) }
| expr MINUSMINUS 
    { mk_expr (Unop (Upost_dec, $1)) }
| expr VERTICALBARVERTICALBAR expr          
    { mk_expr (Binop (Bor, $1, $3)) }
| expr AMPERSANDAMPERSAND expr          
    { mk_expr (Binop (Band, $1, $3)) }
| expr EQOP expr          
    { mk_expr (Binop ($2, $1, $3)) }
| expr COMP expr         
    { mk_expr (Binop ($2, $1, $3)) }
| expr PLUS expr          
    { mk_expr (Binop (Badd, $1, $3)) }
| expr MINUS expr         
    { mk_expr (Binop (Bsub, $1, $3)) }
| expr STAR expr        
    { mk_expr (Binop (Bmul, $1, $3)) }
| expr SLASH expr        
    { mk_expr (Binop (Bdiv, $1, $3)) }
| expr PERCENT expr        
    { mk_expr (Binop (Bmod, $1, $3)) }
| PLUS expr %prec UPLUS 
    { mk_expr (Unop (Uplus, $2)) }
| MINUS expr %prec UMINUS 
    { mk_expr (Unop (Uminus, $2)) }
| BANG expr
    { mk_expr (Unop (Unot, $2)) }
| expr INSTANCEOF type_expr 
    { mk_expr (Instanceof ($1,$3)) } 
| INTEGER                    
    { mk_expr (Constant (Cint $1)) }
| REAL                    
    { mk_expr (Constant (Cdouble $1)) }
| TRUE                       
    { mk_expr (Constant (Cboolean true)) }
| FALSE                      
    { mk_expr (Constant (Cboolean false) )} 
| STRING                     
    { mk_expr (Constant (Cstring $1)) }
| NULL                     
    { mk_expr (Null) }
| THIS
    { mk_expr (This) }
| NEW ident LEFTPAR argument_list RIGHTPAR
    { mk_expr (New ($2, None, $4)) }
| ident LEFTPAR argument_list RIGHTPAR
    { mk_expr (Method_call (None, Cunknown, $1, $3)) }
| expr DOT ident LEFTPAR argument_list RIGHTPAR
    { mk_expr (Method_call (Some $1, Cunknown, $3, $5)) }
| left_value
    { $1 }
/*
  Le transtypage est la seule expression  vraiment poser problme,
  car [(X)-e] et [(X)+e] peuvent autant tre des moins (resp. plus) unaires
  qu'un cast. Heureusement, un moins unaire (ou un plus unaire) ne peut tre 
  transtyp que vers un type de base. 

  Si [X] est un type de base, on lui donne la priorit d'un transtypage.
  Si [X] est un identificateur, on associe la prcdence [par_expr]  [(X)]
  pour que [(X)-e] (resp. [(X)+e]) soit une soustraction (resp. addition).
*/
| LEFTPAR base_type RIGHTPAR expr %prec CAST
    { mk_expr (Cast ($2, $4)) }
| LEFTPAR expr RIGHTPAR expr %prec CAST
    { match $2.expr_node with
	| Field (None, _, c) -> mk_expr (Cast (Tclass c, $4))
	| _ -> 
	    (* [$2] n'est pas une [type_expr] donc l'expression n'est pas dans
	       la grammaire. *)
	    raise Parse_error }
| LEFTPAR expr RIGHTPAR  %prec par_expr
    { $2 }
;

/* valeurs gauches */

left_value:
| ident
    { mk_expr (Field (None, Cunknown, $1)) }
| expr DOT ident
    { mk_expr (Field (Some $1, Cunknown, $3)) }
;

/* Instructions */

expr_comma_list:
| expr
    { [$1] }
| expr COMMA expr_comma_list
    { $1::$3 }
;

statements:
| statement statements
    { $1 :: $2 } 
| /* epsilon */
    { [] } 
;

/* variables locales (n'ont pas de modificateurs) */

local_variable_declaration:
| type_expr variable_declarators
    { { variables_modifiers = no_modifiers ;
	variables_type = $1 ;
	variables_decls = $2 } } 
;

statement:
| expr SEMICOLON
    { mk_st (Expr $1) }
| SEMICOLON
    { mk_st (Skip) }
| local_variable_declaration SEMICOLON
    { mk_st (Var $1) }
| IF LEFTPAR expr RIGHTPAR statement %prec THEN
    { mk_st (If ($3, $5, { st_node = Skip; st_loc = Location.dummy })) }
| IF LEFTPAR expr RIGHTPAR statement ELSE statement
    { mk_st (If ($3, $5, $7)) }
| WHILE LEFTPAR expr RIGHTPAR statement
    { mk_st (While ($3, $5)) }
| FOR LEFTPAR statement_list SEMICOLON expr_option SEMICOLON 
  statement_list RIGHTPAR statement 
    { mk_st (For ($3, $5, $7, $9)) }
| FOR LEFTPAR local_variable_declaration SEMICOLON expr_option SEMICOLON 
  statement_list RIGHTPAR statement 
    { mk_st (Block [ mk_st_i 3 (Var $3); mk_st (For ([], $5, $7, $9)) ]) }
| LEFTBRACE statements RIGHTBRACE
    { mk_st (Block $2) }
| RETURN SEMICOLON
    { mk_st (Return None) }
| RETURN expr SEMICOLON
    { mk_st (Return (Some $2)) }
;

expr_option:
| /* epsilon */ { mk_expr (Constant (Cboolean true)) }
| expr          { $1 }
;

statement_list:
| /* epsilon */ { [] }
| statement_expr_list { $1 }
;

statement_expr_list:
| expr
    { [ mk_st (Expr $1) ] }
| expr COMMA statement_expr_list
    { mk_st (Expr $1) :: $3 }
;

/* paramtres */

method_parameters:
| LEFTPAR RIGHTPAR 
    { [] }
| LEFTPAR parameter_comma_list RIGHTPAR
    { $2 }
;

parameter_comma_list:
| type_expr ident 
    { [ ($1,$2) ] }
| type_expr ident COMMA parameter_comma_list
    { ($1,$2) :: $4 }
;

/* Types. on ne met pas [ident] dans [base_type], pour permettre
   une fois encore la factorisation au niveau des rgles pour 
   le cast. */

base_type:
| BOOLEAN 
    { Tboolean }
| INT 
    { Tint }
| DOUBLE 
    { Tdouble }
;

type_expr:
| ident 
    { Tclass $1 }
| base_type  
    { $1 }
;

/* Modificateurs. On ne construit pas de liste, mais 
   directement le record qui en est l'interprtation. */

modifiers:
| /* epsilon */
    { no_modifiers }
| STATIC modifiers
    { { $2 with m_static = true } }
| PRIVATE modifiers
    { { $2 with m_private = true } }
;

/* identificateurs localiss */

ident:
| IDENT { { id_name = $1; id_loc = loc () } }
;
