
(* Analyse smantique *)

open Format
open Ast

(* Erreurs smantiques *)

type error =
  | ClassClash of string
  | CyclicInheritance
  | UnboundClass of string
  | VarClash of string
  | ConstrName of string
  | ConstrClash
  | MethodClash
  | UnboundVar of string
  | UnboundField of string
  | UnboundMethod of string
  | ExpectedType of string type_expr * string type_expr
  | NoMatch of string type_expr list
  | Ambiguity of string type_expr list
  | NoReturn
  | Private of string * string * string
  | Static of string
  | CastFailure of string type_expr * string type_expr
  | Other of string

let rec pp_type fmt = function
  | Tvoid -> fprintf fmt "void"
  | Tboolean -> fprintf fmt "boolean"
  | Tint -> fprintf fmt "int"
  | Tdouble -> fprintf fmt "double"
  | Tclass c -> fprintf fmt "%s" c
  | Tnull -> fprintf fmt "<null>"

let rec pp_profile fmt = function
  | [] -> ()
  | [t] -> pp_type fmt t
  | t :: p -> fprintf fmt "%a, %a" pp_type t pp_profile p

let report fmt = function
  | ClassClash s -> fprintf fmt "class %s already defined" s
  | CyclicInheritance -> fprintf fmt "cyclic class inheritance"
  | UnboundClass s -> fprintf fmt "unbound class %s" s
  | VarClash s -> fprintf fmt "duplicate variable declaration %s" s
  | ConstrName s -> fprintf fmt "constructor for class %s excepted" s
  | ConstrClash -> fprintf fmt "duplicate constructor declaration"
  | MethodClash -> fprintf fmt "duplicate method declaration"
  | UnboundVar s -> fprintf fmt "unbound variable %s" s
  | UnboundField s -> fprintf fmt "unbound field %s" s
  | UnboundMethod s -> fprintf fmt "unbound method %s" s
  | ExpectedType (t1,t2) -> 
      fprintf fmt 
      "this expression has type %a but is expected to have type %a" 
      pp_type t1 pp_type t2
  | NoReturn -> fprintf fmt "return required at end of method"
  | NoMatch p -> fprintf fmt "no method matching profile (%a)" pp_profile p
  | Ambiguity p -> 
      fprintf fmt 
      "ambiguity: several methods matching profile (%a)" pp_profile p
  | Private (f,c,c0) -> 
      fprintf fmt "Field %s in class %s not accessible from class %s" f c c0
  | Static f ->
      fprintf fmt "static reference to non static variable %s" f
  | CastFailure (ty', ty) ->
      fprintf fmt "expression of type %a cannot be coerced to type %a"
	pp_type ty pp_type ty'
  | Other s -> fprintf fmt "%s" s

exception Error of Location.t * error

let error l e = raise (Error (l, e))

(* La relation d'hritage. [super_class c] donne la super-classe de [c]
   et [extends c1 c2] indique si [c1] tend [c2]. *)

let super_t = Hashtbl.create 17
let super_class = Hashtbl.find super_t
let add_extends c1 c2 = Hashtbl.add super_t c1 c2
let () = add_extends "String" "Object"
let rec extends c1 c2 =
  c1 = c2 || 
  (try let sc1 = Hashtbl.find super_t c1 in extends sc1 c2 
   with Not_found -> false)

(* Oprations sur les types *)

type type_ = string type_expr

let rec subtype t1 t2 = match t1, t2 with
  | Tvoid, Tvoid
  | Tboolean, Tboolean
  | Tint, (Tint | Tdouble)
  | Tdouble, Tdouble 
  | Tnull, (Tnull | Tclass _) -> true
  | Tclass c1, Tclass c2 -> extends c1 c2
  | _ -> false

let compatible t1 t2 = subtype t1 t2 || subtype t2 t1

(* Le module suivant implante des ensembles de profiles, mutables,
   qui serviront  la fois pour les constructeurs et les mthodes 
   surcharges.  *)

module Profile = struct

  type profile = { 
    pclass : string;     (* classe *)
    pargs : type_ list;  (* arguments *)
    pmod : modifiers;    (* modificateurs *)
    ptype : type_;       (* type de retour *)
    pname : string       (* nom long de la forme [type1_type2_..._typek] *)
  }

  (* Pour construire les ensembles, les profiles sont compars par leur classe
     [pclass] et leurs type [pargs]; le type de retour [ptype], les 
     modificateurs [pmod] et le nom long [pname] ne sont pas pris en compte. 
     On dtecte ainsi facilement la tentative (illgale) de re-dclarer un
     profile existant (exception [Already]). *)

  module S = Set.Make(
    struct 
      type t = profile
      let compare p1 p2 = compare (p1.pclass,p1.pargs) (p2.pclass,p2.pargs)
    end)
  type set = S.t ref

  exception Already

  let empty () = ref S.empty

  (* construction du nom long *)

  let rec string_of_type = function
    | Tvoid -> assert false
    | Tboolean -> "boolean"
    | Tint -> "int"
    | Tdouble -> "double"
    | Tclass c -> c
    | Tnull -> "null"

  let rec name = function
    | [] -> ""
    | [t] -> string_of_type t
    | t :: tl -> string_of_type t ^ "_" ^ name tl

  let make c al pmod ty =
    { pclass = c; pargs = al; pmod = pmod; ptype = ty; pname = name al }

  (* l'insertion dtecte le clash *)

  let add p s = 
    if S.mem p !s then raise Already; 
    s := S.add p !s

  (* le profile [p] est plus prcis que [p'] *)

  let sub_profile p p' =
    List.length p = List.length p' && List.for_all2 subtype p p'

  (* mme chose, strictement *)

  let strict_sub_profile p p' = p <> p' && sub_profile p p'

  (* profile tendu avec la classe (pour dtecter les ambiguits
     provenant d'une classe plus prcise mais d'un profile moins
     prcis) *)

  let eprofile p = Tclass p.pclass :: p.pargs

  (* limination dans un ensemble de profiles [ps] de tous les profiles
     (tendus) moins prcis que d'autres dans [ps] *)

  let tighten ps =
    S.fold 
      (fun p s -> 
	 let better_than_p q = strict_sub_profile (eprofile q) (eprofile p) in
	 if S.exists better_than_p ps then 
	   S.remove p s 
	 else 
	   s) 
      ps ps

  (* recherche dans l'ensemble de profiles [s] de tous les profiles
     plus prcis que [p] pour la classe [c], la classe courante tant [cc] *)

  let find cc c p s =
    (* on cherche d'abord tous les profiles filtrant [p] et visibles *)
    let filter p' = 
      (c = cc || not p'.pmod.m_private) &&
      sub_profile (Tclass c :: p) (eprofile p') 
    in
    let ps = S.filter filter !s in
    (* puis on limine ceux qui sont moins prcis que d'autres dans [ps] *)
    S.elements (tighten ps)

  (* ajoute  [s1] les profiles de [s2] qui n'y sont pas dj ;
     les profiles privs ne sont ajouts que si [priv = true] *)
  let union s1 s2 priv =
    S.iter
      (fun p ->
	 let visible = priv || not p.pmod.m_private in
	 if not (S.mem p !s1) && visible then add p s1)
      s2
end

(* Contexte de typage. Celui-ci contient le type des variables locales, y
   compris [this], et la classe courante. La mthode [is_static] indique
   si le contexte est statique, en testant la prsence de [this]. *)

module Context = struct
  module Smap = Map.Make(String)
  type t = string * type_ Smap.t (* class courante, variables locales *)
  let empty c = c, Smap.empty
  let current_class (c,_) = c
  let add x t (c,m) = (c, Smap.add x t m)
  let static_flag b (c,m) = 
    if b then (c, Smap.remove "this" m) else (c, Smap.add "this" (Tclass c) m)
  let find x (_,m) = Smap.find x m
  let mem x (_,m) = Smap.mem x m
  let is_static (_,m) = not (Smap.mem "this" m)
end

(* Table des symboles.

    chaque classe est associe un triplet de tables [(a,c,m)] contenant
   respectivement ses attributs (table de hachage [a] donnant le type et les 
   modificateurs), ses constructeurs (ensemble de profiles [c]) et ses
   mthodes (table de hachage donnant pour chaque nom de mthode un ensemble
   de profiles). *)

let classes = Hashtbl.create 17
let empty_class () = 
  ((Hashtbl.create 17 : (string, type_ * modifiers) Hashtbl.t), 
   Profile.empty (), 
   (Hashtbl.create 17 : (string, Profile.set) Hashtbl.t))
let add_class c = Hashtbl.add classes c (empty_class ())
let () = add_class "Object"
let () = add_class "String"
let is_class = Hashtbl.mem classes
let find_class = Hashtbl.find classes

(* Ajout d'un attribut *)

let add_field c t {id_name=x; id_loc=l} =
  let (f,_,_) = find_class c in
  if Hashtbl.mem f x then error l (VarClash x);
  Hashtbl.add f x t

(* Recherche d'un attribut [x] dans la classe [c]. Le boolen [static] indique
   si l'attribut doit ncessairement tre statique. *)

let rec find_field_type assign loc ctx static c x = 
  let cc = Context.current_class ctx in
  if c = "Object" then raise Not_found;
  let (f,_,_) = find_class c in 
  try 
    let ty,m = Hashtbl.find f x in
    if c <> cc && m.m_private then error loc (Private (x,c,cc));
    if static && not m.m_static then error loc (Static x);
    ty, m.m_static
  with Not_found -> 
    find_field_type assign loc ctx static (super_class c) x

(* Ajout d'un constructeur *)

let add_constructor c l p =
  let (_,ct,_) = find_class c in 
  try Profile.add p ct
  with Profile.Already -> error l ConstrClash

(* Ajout d'une mthode *)

let add_method c {id_name=x; id_loc=l} p =
  let (_,_,mt) = find_class c in
  let px = 
    try Hashtbl.find mt x
    with Not_found -> let px = Profile.empty () in Hashtbl.add mt x px; px
  in
  try Profile.add p px
  with Profile.Already -> error l MethodClash

(* Recherche l'ensemble de toutes les mthodes de nom [m] pour la classe [c0];
   [static] indique si la mthode doit ncessairement tre statique *)

let find_methods loc ctx static c0 m =
  let static_methods = 
    Profile.S.filter (fun p -> p.Profile.pmod.m_static || not static)
  in
  let rec find ms c =
    if c = "Object" then 
      ms
    else begin
      let (_,_,mc) = find_class c in
      (try 
	 let mm = Hashtbl.find mc m in 
	 let mm = static_methods !mm in
	 Profile.union ms mm (c = c0)
       with Not_found -> ());
      find ms (super_class c)
    end
  in
  find (Profile.empty ()) c0

(*s Premire passe : on collecte les classes *)

let declare_class c =
  let id = c.class_id in
  if is_class id.id_name then error id.id_loc (ClassClash id.id_name);
  add_class id.id_name

let phase_1 = List.iter declare_class

(*s Deuxime passe : on interprte les "extends", et on dclare 
    les champs, les constructeurs et les mthodes dans la table des symboles
    (en vrifiant les types associs) *)

let declare_extends c1 = function
  | None -> 
      add_extends c1 "Object"
  | Some { id_name = "String"; id_loc = loc } -> 
      error loc (Other "cannot inherit fromm class String")
  | Some { id_name = c2; id_loc = loc } -> 
      if not (is_class c2) then error loc (UnboundClass c2);
      if extends c2 c1 then error loc CyclicInheritance;
      add_extends c1 c2

let option_app f = function None -> None | Some x -> Some (f x)

let unloc_id id = id.id_name
let unloc_extends = option_app unloc_id

(* Bonne formation d'un type (retourne le meme type, dans lequel les
   identificateurs ont perdu leurs positions) *)

let check_type = function
  | Tclass { id_name = c; id_loc = l } -> 
      if not (is_class c) then error l (UnboundClass c);
      Tclass c
  | Tvoid | Tboolean | Tint | Tdouble | Tnull as t ->
      t

(* Bonne formation des paramtres d'un constructeur ou d'une mthode :
   vrifie que les noms des paramtres sont diffrents deux  deux et 
   que le type de chaque paramtre est bien form.
   Retourne la liste des types des paramtres *)

let check_parameters params = 
  let check_name names id =
    let n = id.id_name in
    if List.exists ((=) n) names then error id.id_loc (VarClash n);
    n :: names
  in
  fst (List.fold_right
	 (fun (t, n) (tl, nl) -> check_type t :: tl, check_name nl n)
	 params
	 ([], []))

let type_parameter (ty, id) = (check_type ty, unloc_id id)
let type_parameters = List.map type_parameter

(* Dclaration des variables d'instance *)

let declare_variables c vd =
  let ty = check_type vd.variables_type, vd.variables_modifiers in
  List.iter (fun v -> add_field c ty v.variable_id) vd.variables_decls

(* Dclaration des constructeurs *)

let declare_constructor c cd =
  let l = cd.constr_name.id_loc in
  let id = cd.constr_name.id_name in
  if id <> c then error l (ConstrName c);
  let cmod = cd.constr_modifiers in
  if cmod.m_static then 
    error cd.constr_name.id_loc 
      (Other "a constructor can't be declared static");
  let al = check_parameters cd.constr_parameters in
  add_constructor c l (Profile.make c al cmod Tvoid)

(* Dclaration des mthodes *)

let declare_method c md =
  let mmod = md.method_modifiers in
  let al = check_parameters md.method_parameters in
  let ty = check_type md.method_type in
  add_method c md.method_name (Profile.make c al mmod ty)

let declare_fields c = 
  let id = c.class_id.id_name in
  declare_extends id c.class_extends;
  List.iter
    (function
       | Variable vd -> declare_variables id vd
       | Method md -> declare_method id md
       | Constructor cd -> declare_constructor id cd)
    c.class_fields

let phase_2 = List.iter declare_fields

(*s Troisime passe : typage proprement dit. *)

(* Coercion explicite de [int] vers [double] *)

let double e = { e with expr_node = Unop (Udouble_of_int, e) }
let truncate e = { e with expr_node = Unop (Uint_of_double, e) }

let cast t1 t2 e = if t1 = Tint && t2 = Tdouble then double e else e

let coerce loc t1 t2 e =
  if not (subtype t1 t2) then error loc (ExpectedType (t1, t2));
  cast t1 t2 e

let rec coerce_list e t et = match e, t, et with
  | [], [], [] ->
      []
  | e :: el, Tint :: tyl, Tdouble :: etyl ->
      double e :: coerce_list el tyl etyl
  | e :: el, ty :: tyl, ety :: etyl ->
      assert (subtype ty ety);
      e :: coerce_list el tyl etyl
  | _ ->
      assert false

(* La classe [String] *)

let class_String = Tclass "String"

(* Oprations sur les flottants *)

let double_op = function
  | Badd -> Badd_f
  | Bsub -> Bsub_f
  | Bmul -> Bmul_f
  | Bdiv -> Bdiv_f
  | Blt  -> Blt_f
  | Ble  -> Ble_f
  | Bgt  -> Bgt_f
  | Bge  -> Bge_f
  | _ -> assert false

(* [class_info c private static] construit une information de classe pour la
   classe [c] et le caractre statique [static]. *)

let class_info c p = function true -> Cstatic c | false -> Cdynamic (c, p)

(* Recherche d'un profile unique dans l'ensemble de profile [pm] pour
   une liste d'arguments [tel] de types [tyl]. Si aucun profile n'est trouv,
   c'est une erreur. Si plusieurs profiles sont trouvs, c'est galement
   une erreur smantique (ambiguit). *)

let find_profile ctx loc c (tel,tyl) pm = 
  let cc = Context.current_class ctx in
  match Profile.find cc c tyl pm with
    | [] -> 
	error loc (NoMatch tyl)
    | [p] -> 
	let etyl = p.Profile.pargs in
	coerce_list tel tyl etyl, p
    | _ -> 
      error loc (Ambiguity tyl)

(* Typage des expressions *)

let rec type_expr ctx e = 
  let loc = e.expr_loc in
  let te,ty = type_expr_node ctx loc e.expr_node in
  { expr_node = te; expr_loc = loc }, ty

and type_expr_node ctx loc = function
  | Constant (Cint _) as c -> 
      c, Tint
  | Constant (Cdouble _) as c -> 
      c, Tdouble
  | Constant (Cstring _) as c -> 
      c, class_String
  | Constant (Cboolean _) as c -> 
      c, Tboolean
  | Null -> 
      Null, Tnull
  | This -> 
      This,
      (try Context.find "this" ctx 
       with Not_found -> error loc (UnboundVar "this"))
  | Field _ as e ->
      type_left_value_node false loc ctx e
  (* System.out.println(e) *)
  | Method_call 
      (Some {expr_node=
	       Field (Some {expr_node = Field (None, _, {id_name="System"})},
		      _, {id_name="out"})},
       _, {id_name="println"},
       [e]) ->
      (match type_expr ctx e with
	 | te, Tclass "String" -> Println te, Tvoid
	 | _, ty -> error loc (ExpectedType (ty, class_String)))
  | Method_call (ob, _, {id_name=m}, el) ->
      let tob,c,static = match ob with
	| None -> 
	    None, Context.current_class ctx, Context.is_static ctx
	| Some ({expr_node=Field(None,_,{id_name=c})} as e) when is_class c ->
	    None, c, true
	| Some e -> match type_expr ctx e with
	    | te, Tclass c -> Some te, c, false
	    | _ -> error e.expr_loc (Other "this expression is not an object")
      in
      if not (is_class c) then error loc (UnboundClass c); (* utile ? *)
      let tel = type_expr_l ctx el in
      let pm = find_methods loc ctx static c m in
      (try
	 let tel,p = find_profile ctx loc c tel pm in
	 let pmod = p.Profile.pmod in
	 Method_call 
	   (tob, 
	    class_info p.Profile.pclass pmod.m_private pmod.m_static, 
	    m ^ "_" ^ p.Profile.pname, tel), 
	 p.Profile.ptype
       with Error(_, NoMatch _) as exn ->
	 match tel with
	   | te, [ Tclass "String" ] -> 
	       if m = "equals" && c = "String" then
		 (* mthode "equals" des String. *)
		 let mmod = { m_static = false; m_private = false } in
		 let p = 
		   Profile.make "String" [Tclass "String"] mmod Tboolean 
		 in
		 Method_call (tob, class_info "String" false false, 
			      m ^ "_" ^ p.Profile.pname, te),
		 Tboolean
	       else raise exn
	   | _     -> raise exn)
  | New ({id_name=c}, _, el) ->
      let (_,ct,_) = 
	try find_class c with Not_found -> error loc (UnboundClass c)
      in
      let tel = type_expr_l ctx el in
      let tel,p = find_profile ctx loc c tel ct in
      New (c, Some p.Profile.pname, tel), Tclass c
  | Assign (e1, e2) ->
      let te1,ty1 = type_left_value true ctx e1 in
      let te2,ty2 = type_expr ctx e2 in
      let te2 = coerce e2.expr_loc ty2 ty1 te2 in
      Assign (te1, te2), ty1
  | Binop (op, e1, e2) ->
      let te1,ty1 = type_expr ctx e1 in
      let te2,ty2 = type_expr ctx e2 in
      let stri e = { e with expr_node = Unop (Ustring_of_int, e) } in
      let strf e = { e with expr_node = Unop (Ustring_of_float, e) } in
      (match op, ty1, ty2 with
	| (Beq | Bneq), Tint, Tdouble  -> 
	    Binop (op, double te1, te2), Tboolean
	| (Beq | Bneq), Tdouble, Tint -> 
	    Binop (op, te1, double te2), Tboolean
	| (Beq | Bneq), _, _ when compatible ty1 ty2 -> 
	    Binop (op, te1, te2), Tboolean
	| (Blt | Ble | Bgt | Bge), Tint, Tint -> 
	    Binop (op, te1, te2), Tboolean
	| (Blt | Ble | Bgt | Bge), Tint, Tdouble -> 
	    Binop (double_op op, double te1, te2), Tboolean
	| (Blt | Ble | Bgt | Bge), Tdouble, Tint -> 
	    Binop (double_op op, te1, double te2), Tboolean
	| (Blt | Ble | Bgt | Bge), Tdouble, Tdouble -> 
	    Binop (double_op op, te1, te2), Tboolean
	| (Badd | Bsub | Bmul | Bdiv), Tint, Tint -> 
	    Binop (op, te1, te2), Tint
	| (Badd | Bsub | Bmul | Bdiv), Tint, Tdouble -> 
	    Binop (double_op op, double te1, te2), Tdouble
	| (Badd | Bsub | Bmul | Bdiv), Tdouble, Tint -> 
	    Binop (double_op op, te1, double te2), Tdouble
	| (Badd | Bsub | Bmul | Bdiv), Tdouble, Tdouble -> 
	    Binop (double_op op, te1, te2), Tdouble
	| Bmod, Tint, Tint -> 
	    Binop (op, te1, te2), Tint
	| (Band | Bor), Tboolean, Tboolean -> 
	    Binop (op, te1, te2) ,Tboolean
	| Badd, Tclass "String", Tclass "String" ->
	    Binop (Badd_s, te1, te2), class_String
	| Badd, Tclass "String", Tint ->
	    Binop (Badd_s, te1, stri te2), class_String
	| Badd, Tclass "String", Tdouble ->
	    Binop (Badd_s, te1, strf te2), class_String
	| Badd, Tint, Tclass "String" ->
	    Binop (Badd_s, stri te1, te2), class_String
	| Badd, Tdouble, Tclass "String" ->
	    Binop (Badd_s, strf te1, te2), class_String
        (* cas d'erreurs *)
	| (Beq | Bneq), _, _ -> error loc (Other "incompatible types")
	| (Blt | Ble | Bgt | Bge | Badd | Bsub|Bmul|Bdiv), (Tint|Tdouble), _ ->
	    error e2.expr_loc (ExpectedType (ty2, Tclass "int or double"))
	| Badd, Tclass "String", _ ->
	    error e2.expr_loc (ExpectedType (ty2, 
					     Tclass "String, int or double"))
	| Badd, _, Tclass "String" ->
	    error e1.expr_loc (ExpectedType (ty1, 
					     Tclass "String, int or double"))
	| (Blt | Ble | Bgt | Bge | Badd | Bsub | Bmul | Bdiv), _, _ ->
	    error e1.expr_loc (ExpectedType (ty1, Tclass "int or double"))
 	| Bmod, Tint, _ ->
	    error e2.expr_loc (ExpectedType (ty2, Tint))
 	| Bmod, _, _ ->
	    error e1.expr_loc (ExpectedType (ty1, Tint))
	| (Band | Bor), Tboolean, _ -> 
	    error e2.expr_loc (ExpectedType (ty2, Tboolean))
	| (Band | Bor), _, _ -> 
	    error e1.expr_loc (ExpectedType (ty1, Tboolean))
        (* impossible *)
	| (Bdiv_f|Bmul_f|Bsub_f|Badd_f|Bge_f|Bgt_f|Ble_f|Blt_f|Badd_s), _, _ ->
	    assert false)
  | Unop (Upre_inc | Upost_inc | Upre_dec | Upost_dec as op, e) ->
      (match type_left_value true ctx e with
	 | te, Tint -> Unop (op, te), Tint
	 | _, ty -> error e.expr_loc (ExpectedType (ty, Tint)))
  | Unop (Unot, e) ->
      (match type_expr ctx e with
	 | te, Tboolean -> Unop (Unot, te), Tboolean
	 | _, ty -> error e.expr_loc (ExpectedType (ty, Tboolean)))	 
  | Unop (Uminus, e) ->
      (match type_expr ctx e with
	 | te, Tint -> Unop (Uminus, te), Tint
	 | te, Tdouble -> Unop (Uminus_f, te), Tdouble
	 | _, ty -> 
	     error e.expr_loc (ExpectedType (ty, Tclass "int or double")))
  | Unop (Uplus, e) ->
      (match type_expr ctx e with
	 | te, (Tint | Tdouble as t) -> 
	     te.expr_node, t (* on peut jeter l'oprateur *)
	 | _, ty -> 
	     error e.expr_loc (ExpectedType (ty, Tclass "int or double")))
  | Unop ((Udouble_of_int | Uint_of_double | 
	   Uminus_f | Ustring_of_int | Ustring_of_float), _) ->
      assert false
  | Cast (ty, e) -> 
      let ty = check_type ty in
      let te,ty' = type_expr ctx e in
      (match ty, ty' with
	 | Tint, Tdouble -> Unop (Uint_of_double, te)
	 | Tdouble, Tint -> Unop (Udouble_of_int, te)
	 | _ when ty = ty' -> te.expr_node
	 | Tclass _, (Tclass _ | Tnull) ->
	     if subtype ty' ty then te.expr_node 
	     else if subtype ty ty' then Cast (ty, te)
	     else error loc (CastFailure (ty', ty))
	 | _ ->
	     error loc (CastFailure (ty', ty)))
      , ty
  | Instanceof (e, (Tclass _ as ty)) ->
      let ty = check_type ty in
      let te,ty' = type_expr ctx e in
      (match ty', ty with
	 | (Tint | Tdouble | Tboolean | Tvoid), _ ->
	     error loc (Other "inconvertible types")
	 | Tnull, _ ->
	     Constant (Cboolean false)
	 | _ when compatible ty ty' ->
	     Instanceof (te, ty)
	 | _ ->
	     error loc (Other "inconvertible types")),
      Tboolean
  | Instanceof _ ->
      error loc (Other "instanceof expects a class")
  | Println _ ->
      assert false

(* Typage des valeurs gauches ([Field]). *)

and type_left_value assign ctx e =
  let te,ty = type_left_value_node assign e.expr_loc ctx e.expr_node in
  { e with expr_node = te }, ty

and type_left_value_node assign loc ctx = function
  | Field (Some { expr_node = Field (None,_,{id_name=c}) }, _, {id_name=x}) 
    when is_class c ->
      (try
	 let ty,s = find_field_type assign loc ctx true c x in
	 if not s then error loc (Static x);
	 Field (None, Cstatic c, x), ty
       with Not_found -> error loc (UnboundField x))
  (* Note: le caractre priv ne joue pas pour les attributs *)
  | Field (None, _, {id_name=x}) as e ->
      (try 
	 (* variable locale ? *)
	 let ty = Context.find x ctx in Field (None, Clocal, x), ty
       with Not_found -> 
	 (* champ pour la classe courante ? *)
	 let c = Context.current_class ctx in
	 try 
	   let static = Context.is_static ctx in
	   let ty,s = find_field_type assign loc ctx static c x in
	   Field (None, class_info c false s, x), ty
	 with Not_found -> error loc (UnboundVar x))
  | Field (Some e, _, {id_name=x}) -> 
      (try
	 (match type_expr ctx e with
	    | te, Tclass c -> 
		let ty,s = find_field_type assign loc ctx false c x in
		Field (Some te, class_info c false s, x), ty
	    | _ -> 
		error e.expr_loc (Other "this expression is not an object"))
       with Not_found -> error loc (UnboundField x))
  | _ -> 
      error loc (Other "not a left value")

(* Typage d'une liste d'arguments. *)

and type_expr_l ctx = function
  | [] -> 
      [], []
  | e :: el -> 
      let te,ty = type_expr ctx e in
      let tel,tyl = type_expr_l ctx el in
      te :: tel, ty :: tyl

(* Typage des instructions.

   [et] est le type (option) attendu de la valeur de retour;
   en retour, on retourne le nouveau contexte, 
   et on indique si l'instruction retourne dans tous les flots de
   contrle possibles. *)

let rec type_statement ctx et st = 
  let loc = st.st_loc in
  let tst,ctx',r = type_statement_node ctx et loc st.st_node in
  { st_node = tst; st_loc = loc }, ctx', r

and type_statement_node ctx et loc = function
  | Skip -> 
      Skip, ctx, false
  | Expr e ->
      let te,_ = type_expr ctx e in
      Expr te, ctx, false
  | Var ({ variables_type = ty; variables_decls = vl } as vd) -> 
      let ty = check_type ty in
      let vl',ctx' = type_local_variables ctx ty vl in
      Var { vd with variables_type = ty; variables_decls = vl' }, ctx', false
  | If (e1, st2, st3) ->
      let te1 = match type_expr ctx e1 with
	| te1, Tboolean -> te1
	| _, ty -> error e1.expr_loc (ExpectedType (ty, Tboolean))
      in
      let tst2,_,r2 = type_statement ctx et st2 in
      let tst3,_,r3 = type_statement ctx et st3 in
      If (te1, tst2, tst3), ctx, r2 && r3
  | While (e, st) ->
      let te = match type_expr ctx e with
	| te, Tboolean -> te
	| _, ty -> error e.expr_loc (ExpectedType (ty, Tboolean))
      in
      let tst,_,_ = type_statement ctx et st in
      While (te, tst), ctx, false
  | For (bl1, e2, bl3, st4) ->
      let tbl1,ctx',_ = type_block ctx et bl1 in
      let te2 = match type_expr ctx' e2 with
	| te2, Tboolean -> te2
	| _, ty -> error e2.expr_loc (ExpectedType (ty, Tboolean))
      in
      let tbl3,_,_ = type_block ctx' et bl3 in
      let tst4,_,_ = type_statement ctx' et st4 in
      For (tbl1, te2, tbl3, tst4), ctx, false
  | Block bl ->
      let tbl,_,r = type_block ctx et bl in
      Block tbl, ctx, r
  | Return e ->
      let te,ty = match e with
	| None -> 
	    if et = Tvoid then 
	      None, Tvoid
	    else 
	      error loc (ExpectedType (Tvoid, et))
	| Some e -> 
	    let te,ty = type_expr ctx e in 
	    let te = coerce loc ty et te in
	      Some te, ty
      in
      Return te, ctx, true

(* Typage d'un bloc. *)

and type_block ctx et = function
  | [] -> 
      [], ctx, false
  | st :: bl ->
      let tst,ctx',rst = type_statement ctx et st in
      let tbl,ctx'',rbl = type_block ctx' et bl in
      tst :: tbl, ctx'',rst || rbl

(* Typage d'une dclaration de variables locales. *)

and type_local_variables ctx ty vl =
  let ctx',vl' = 
    List.fold_left
      (fun (ctx',vl') {variable_id=v; variable_initializer=e} ->
	 let x = v.id_name in
	 if Context.mem x ctx' then error v.id_loc (VarClash x);
	 let te = match e with
	   | None -> 
	       None
	   | Some e -> 
	       let te,tye = type_expr ctx' e in
	       let te = coerce e.expr_loc tye ty te in
	       Some te
	 in
	 let v' = { variable_id = x; variable_initializer = te } in
	 Context.add x ty ctx', v' :: vl')
      (ctx,[]) vl
  in
  List.rev vl', ctx'

(* Typage des initialisations des attributs. *)

let type_var ctx vd =
  let ty = check_type vd.variables_type in
  let ctx = Context.static_flag vd.variables_modifiers.m_static ctx in
  let vl = vd.variables_decls in
  let vl',_ = type_local_variables ctx ty vl in
  { vd with variables_type = ty; variables_decls = vl' }

(* Typage des constructeurs. *)

let type_constr ctx cd =
  let id = cd.constr_name.id_name in
  let loc = cd.constr_name.id_loc in
  let cc = Context.current_class ctx in
  if id <> cc then error loc (Other "return type expected");
  let ctx = Context.static_flag false ctx in
  let args = type_parameters cd.constr_parameters in
  let ctx = List.fold_left (fun ctx (ty,x) -> Context.add x ty ctx) ctx args in
  let targs = List.map fst args in
  let body,_,_ = type_block ctx Tvoid cd.constr_body in
  { cd with constr_name = id ^ "_" ^ Profile.name targs;
            constr_parameters = args;
	    constr_body = body }

(* Typage des mthodes. *)

(* Ici on vrifie que [md] n'est pas redfinie 
   (1) comme plus prive,
   (2) avec un type de retour diffrent, 
   (3) avec un caractre statique diffrent *)

let check_override loc c0 md =
  let m = md.method_name in
  let args = List.map fst md.method_parameters in
  let mods = md.method_modifiers in
  let check_profile p =
    let pmods = p.Profile.pmod in
    if p.Profile.pargs = args then begin
      if mods.m_static <> pmods.m_static then
	error loc (Other ("a static method can't be overridden by a non-static"
			  ^ ", and vice-versa"));
      if p.Profile.ptype <> md.method_type then 
	error loc 
	  (Other "a method can't be overridden with a different type");
      if mods.m_private && not pmods.m_private then
	error loc (Other "a method can't be overridden to be more private")
    end
  in
  let rec check c =
    if c <> "Object" then begin
      let (_,_,mt) = find_class c in
      begin try
	let ps = Hashtbl.find mt m in
	Profile.S.iter check_profile !ps
      with Not_found -> () end;
      check (super_class c)
    end
  in
  check (super_class c0)

let type_method ctx md =
  let m = unloc_id md.method_name in
  let mods = md.method_modifiers in
  let ty = check_type md.method_type in
  let ctx = Context.static_flag mods.m_static ctx in
  let args = type_parameters md.method_parameters in
  let targs = List.map fst args in
  let ctx = List.fold_left (fun ctx (ty,x) -> Context.add x ty ctx) ctx args in
  let b,_,r = type_block ctx ty md.method_body in
  if ty <> Tvoid && not r then error md.method_name.id_loc NoReturn;
  let md' = { md with
		method_type = ty;
		method_name = m;
		method_parameters = args;
		method_body = b } in
  check_override md.method_name.id_loc (Context.current_class ctx) md';
  { md' with method_name = m ^ "_" ^ Profile.name targs }

(* Typage des classes. *)

let type_field ctx = function
  | Variable vd -> Variable (type_var ctx vd)
  | Constructor cd -> Constructor (type_constr ctx cd)
  | Method md -> Method (type_method ctx md)

let type_fields c =
  let id = c.class_id.id_name in
  let ctx = Context.empty id in
  { class_id = id;
    class_extends = unloc_extends c.class_extends;
    class_fields = List.map (type_field ctx) c.class_fields }

let phase_3 = List.map type_fields

(* Fonction de typage : effectue les 3 passes, la troisime retournant
   le fichier typ pour la production de code  venir. *)

let compilation_unit cu  = 
  phase_1 cu;
  phase_2 cu;
  phase_3 cu

(* Vrification du point d'entre (appel par [Compile]). *)

let check_entry_point (c,m) =
  let loc = Location.dummy in
  let ctx = Context.empty c in
  try
    let pm = find_methods loc ctx true c m in
    let _ = find_profile ctx loc c ([],[]) pm in
    ()
  with _ ->
    error loc (Other "entry point must be a static method C.m()")
